#include <stdio.h>
#include <string.h>
#include <vdpau/vdpau_x11.h>
#include <cuda/cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_vdpau_interop.h>
#include "traceplay.h"
#include "traceplay_cuda_kernels.cu"
#include "common.h"

typedef struct {
    struct cudaGraphicsResource *gr;
} RegisteredVideoSurface;

typedef struct {
    struct cudaGraphicsResource *gr;
} RegisteredOutputSurface;

Display *     x_display;
int           x_screen;
Window        x_window;

VdpDevice           vdp_device;
VdpGetProcAddress * vdp_get_proc_address;

VdpPresentationQueueTarget vdp_flip_target;
VdpPresentationQueue       vdp_flip_queue;

VdpDeviceDestroy *                          vdp_device_destroy;
VdpPresentationQueueTargetDestroy *         vdp_presentation_queue_target_destroy;
VdpPresentationQueueCreate *                vdp_presentation_queue_create;
VdpPresentationQueueDestroy *               vdp_presentation_queue_destroy;
VdpPresentationQueueDisplay *               vdp_presentation_queue_display;
VdpPresentationQueueBlockUntilSurfaceIdle * vdp_presentation_queue_block_until_surface_idle;
VdpPresentationQueueTargetCreateX11 *       vdp_presentation_queue_target_create_x11;

void *cu_dptr_out, *cu_dptr_weaved_luma, *cu_dptr_weaved_chroma;
size_t pitch_out, pitch_weaved_luma, pitch_weaved_chroma;

void * traceplay_video_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)malloc(sizeof(*rvs));
    CHECK_PTR(rvs)
    
    cudaError_t cr = cudaGraphicsVDPAURegisterVideoSurface(&rvs->gr, surface, CU_GRAPHICS_MAP_RESOURCE_FLAGS_READ_ONLY);
    CHECK_CUDA_ERROR

    return rvs;
}

void * traceplay_output_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface
)
{
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)malloc(sizeof(*ros));
    CHECK_PTR(ros)
    
    cudaError_t cr = cudaGraphicsVDPAURegisterOutputSurface(&ros->gr, surface, CU_GRAPHICS_MAP_RESOURCE_FLAGS_WRITE_DISCARD);
    CHECK_CUDA_ERROR

    return ros;
}

void traceplay_video_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)surface_context;

    cudaError_t cr = cudaGraphicsUnregisterResource(rvs->gr);
    CHECK_CUDA_ERROR

    free(rvs);
}

void traceplay_output_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)surface_context;

    cudaError_t cr = cudaGraphicsUnregisterResource(ros->gr);
    CHECK_CUDA_ERROR

    free(ros);
}

void traceplay_display_video_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    printf("ERROR at %s:%d: display_video_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

void traceplay_wait_for_output_surface_idle(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    VdpTime t;
    vdp_presentation_queue_block_until_surface_idle(vdp_flip_queue, surface, &t);
}

void traceplay_display_output_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    vdp_presentation_queue_display(vdp_flip_queue, surface, 0, 0, 0);
}

void cuda_toolkit_do_gold_interop_vid(uint32_t frame_number, struct cudaArray * arvtl, struct cudaArray * arvbl, struct cudaArray * arvtc, struct cudaArray * arvbc)
{
    int i;
    cudaError_t cr;

    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    uint8_t *buffer = scratch_surface_buffer;
    size_t height = trace_stream_conf_ptr->vid_height / 2;
    size_t texsize = trace_stream_conf_ptr->vid_width * height;
    struct cudaArray * ars[] = {arvtl, arvbl, arvtc, arvbc};

    for (i = 0; i < 4; ++i) {
        cr = cudaMemcpy2DFromArray(buffer, trace_stream_conf_ptr->vid_width, ars[i], 0, 0, trace_stream_conf_ptr->vid_width, height, cudaMemcpyDeviceToHost);
        CHECK_CUDA_ERROR

        buffer += texsize;
        if (i == 1) {
            height /= 2;
            texsize /= 2;
        }
    }

    common_nv12x2_to_nv12(trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, scratch_surface_buffer, surface_buffer);

    common_do_gold_buffer(frame_number, "nv12", trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, surface_buffer, surface_buffer_size);
}

void cuda_toolkit_do_gold_interop_out(uint32_t frame_number, struct cudaArray * aro)
{
    cudaError_t cr;

    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    cr = cudaMemcpy2DFromArray(surface_buffer, trace_stream_conf_ptr->out_width * 4, aro, 0, 0, trace_stream_conf_ptr->out_width * 4, trace_stream_conf_ptr->out_height, cudaMemcpyDeviceToHost);
    CHECK_CUDA_ERROR

    common_do_gold_buffer_argb_to_rgb(frame_number, trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height, surface_buffer, surface_buffer_size, 0);
}

void traceplay_process_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        vid_surface,
    void *                 vid_surface_context,
    VdpOutputSurface       out_surface,
    void *                 out_surface_context
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)vid_surface_context;
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)out_surface_context;
    cudaError_t cr;

    // Map and acquire arrays

    cr = cudaGraphicsMapResources(1, &rvs->gr, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsMapResources(1, &ros->gr, 0);
    CHECK_CUDA_ERROR

    struct cudaArray *arvtl, *arvbl, *arvtc, *arvbc, *aro;

    cr = cudaGraphicsSubResourceGetMappedArray(&arvtl, rvs->gr, 0, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsSubResourceGetMappedArray(&arvbl, rvs->gr, 1, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsSubResourceGetMappedArray(&arvtc, rvs->gr, 2, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsSubResourceGetMappedArray(&arvbc, rvs->gr, 3, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsSubResourceGetMappedArray(&aro, ros->gr, 0, 0);
    CHECK_CUDA_ERROR

    if (opt_gold_surface_type == SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID) {
        cuda_toolkit_do_gold_interop_vid(frame_number, arvtl, arvbl, arvtc, arvbc);
    }

    // Weave luma

    cr = cudaBindTextureToArray(luma_top, arvtl);
    CHECK_CUDA_ERROR
    luma_top.normalized = 0;
    luma_top.filterMode = cudaFilterModePoint;

    cr = cudaBindTextureToArray(luma_bottom, arvbl);
    CHECK_CUDA_ERROR
    luma_bottom.normalized = 0;
    luma_bottom.filterMode = cudaFilterModePoint;

    {
        dim3 dimGrid(trace_stream_conf_ptr->vid_width / 16, trace_stream_conf_ptr->vid_height / 16);
        dim3 dimBlock(16, 16);
        luma_weave<<<dimGrid, dimBlock>>>((unsigned char *)cu_dptr_weaved_luma, pitch_weaved_luma);
    }

    // Weave Chroma

    cr = cudaBindTextureToArray(chroma_top, arvtc);
    CHECK_CUDA_ERROR
    chroma_top.normalized = 0;
    chroma_top.filterMode = cudaFilterModePoint;

    cr = cudaBindTextureToArray(chroma_bottom, arvbc);
    CHECK_CUDA_ERROR
    chroma_bottom.normalized = 0;
    chroma_bottom.filterMode = cudaFilterModePoint;

    {
        dim3 dimGrid(trace_stream_conf_ptr->vid_width / (8 * 2), trace_stream_conf_ptr->vid_height / (8 * 2));
        dim3 dimBlock(8, 8);
        chroma_weave<<<dimGrid, dimBlock>>>((unsigned short *)cu_dptr_weaved_chroma, pitch_weaved_chroma);
    }

    // CSC

    {
        cudaChannelFormatDesc desc = cudaCreateChannelDesc<uchar1>();
        cr = cudaBindTexture2D(0, luma, cu_dptr_weaved_luma, desc, trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, pitch_weaved_luma);
        CHECK_CUDA_ERROR
        luma.normalized = 1;
        luma.filterMode = cudaFilterModeLinear;
    }

    {
        cudaChannelFormatDesc desc = cudaCreateChannelDesc<uchar2>();
        cr = cudaBindTexture2D(0, chroma, cu_dptr_weaved_chroma, desc, trace_stream_conf_ptr->vid_width / 2, trace_stream_conf_ptr->vid_height / 2, pitch_weaved_luma);
        CHECK_CUDA_ERROR
        chroma.normalized = 1;
        chroma.filterMode = cudaFilterModeLinear;
    }

    {
        dim3 dimGrid((trace_stream_conf_ptr->out_width + 15) / 16, (trace_stream_conf_ptr->out_height + 15) / 16);
        dim3 dimBlock(16, 16);
        csc<<<dimGrid, dimBlock>>>((unsigned int *)cu_dptr_out, pitch_out, trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height);
    }

    // Copy results back to VDPAU surface

    cr = cudaMemcpy2DToArray(aro, 0, 0, cu_dptr_out, pitch_out, trace_stream_conf_ptr->out_width * 4, trace_stream_conf_ptr->out_height, cudaMemcpyDeviceToDevice);
    CHECK_CUDA_ERROR

    if (opt_gold_surface_type == SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT) {
        cuda_toolkit_do_gold_interop_out(frame_number, aro);
    }

    // Unmap

    cr = cudaGraphicsUnmapResources(1, &ros->gr, 0);
    CHECK_CUDA_ERROR
    cr = cudaGraphicsUnmapResources(1, &rvs->gr, 0);
    CHECK_CUDA_ERROR
}

Traceplay_Callbacks tpcb = {
    traceplay_video_surface_post_create,
    traceplay_output_surface_post_create,
    traceplay_video_surface_pre_destroy,
    traceplay_output_surface_pre_destroy,
    traceplay_display_video_surface,
    traceplay_wait_for_output_surface_idle,
    traceplay_display_output_surface,
    traceplay_process_surface
};

void initX11(void)
{
    unsigned long windowMask;
    XSetWindowAttributes winAttrib;

    // X11 setup

    x_display = XOpenDisplay(NULL);
    if (!x_display) {
        printf("ERROR at %s:%d: XOpenDisplay failed\n", __FILE__, __LINE__);
        common_exit_error();
    }
    x_screen = DefaultScreen(x_display);

    windowMask = CWBorderPixel; 
    winAttrib.border_pixel = BlackPixel(x_display, x_screen);
    winAttrib.override_redirect = 0;

    x_window = XCreateWindow(
        x_display,
        RootWindow(x_display, x_screen),
        100,
        100,
        trace_stream_conf_ptr->out_width,
        trace_stream_conf_ptr->out_height,
        0,
        CopyFromParent,
        InputOutput,
        CopyFromParent,
        windowMask,
        &winAttrib
    );

    XSelectInput(x_display, x_window, StructureNotifyMask);

    XMapRaised(x_display, x_window);

    XFlush(x_display);
}

void finiX11(void)
{
    XDestroyWindow(x_display, x_window);
    XCloseDisplay(x_display);
}

void initCUDA(void)
{
    cudaError_t cr;
    int cu_dev;

    cr = cudaVDPAUGetDevice(&cu_dev, vdp_device, vdp_get_proc_address);
    CHECK_CUDA_ERROR
    cr = cudaVDPAUSetVDPAUDevice(cu_dev, vdp_device, vdp_get_proc_address);
    CHECK_CUDA_ERROR
    cr = cudaMallocPitch(&cu_dptr_out, &pitch_out, trace_stream_conf_ptr->out_width * 4, trace_stream_conf_ptr->out_height);
    CHECK_CUDA_ERROR
    cr = cudaMallocPitch(&cu_dptr_weaved_luma, &pitch_weaved_luma, trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height);
    CHECK_CUDA_ERROR
    cr = cudaMallocPitch(&cu_dptr_weaved_chroma, &pitch_weaved_chroma, trace_stream_conf_ptr->vid_width, (trace_stream_conf_ptr->vid_height / 2));
    CHECK_CUDA_ERROR
}

void finiCUDA(void)
{
    cudaError_t cr;

    if (cu_dptr_weaved_chroma) {
        cr = cudaFree(cu_dptr_weaved_chroma);
        CHECK_CUDA_ERROR
        cu_dptr_weaved_chroma = 0;
    }

    if (cu_dptr_weaved_luma) {
        cr = cudaFree(cu_dptr_weaved_luma);
        CHECK_CUDA_ERROR
        cu_dptr_weaved_luma = 0;
    }

    if (cu_dptr_out) {
        cr = cudaFree(cu_dptr_out);
        CHECK_CUDA_ERROR
        cu_dptr_out = 0;
    }

    cr = cudaThreadExit();
    CHECK_CUDA_ERROR
}

void initVDPAU(void)
{
    VdpStatus vdp_st;

    // Create Device

    vdp_st = vdp_device_create_x11(
        x_display,
        x_screen,
        &vdp_device,
        &vdp_get_proc_address
    );
    CHECK_ST

    initCUDA();

    // Entry point retrieval

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DEVICE_DESTROY,
        (void **)&vdp_device_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_DESTROY,
        (void **)&vdp_presentation_queue_target_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_CREATE,
        (void **)&vdp_presentation_queue_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DESTROY,
        (void **)&vdp_presentation_queue_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DISPLAY,
        (void **)&vdp_presentation_queue_display
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_BLOCK_UNTIL_SURFACE_IDLE,
        (void **)&vdp_presentation_queue_block_until_surface_idle
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_CREATE_X11,
        (void **)&vdp_presentation_queue_target_create_x11
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_target_create_x11(
        vdp_device,
        x_window,
        &vdp_flip_target
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_create(
        vdp_device,
        vdp_flip_target,
        &vdp_flip_queue
    );
    CHECK_ST
}

void finiVDPAU(void)
{
    vdp_presentation_queue_destroy(
        vdp_flip_queue
    );

    vdp_presentation_queue_target_destroy(
        vdp_flip_target
    );

    vdp_device_destroy(
        vdp_device
    );
}

void help()
{
    printf("app [options]\n");
    common_print_help();
}

int main(int argc, char * * argv)
{
    cudaError_t cr;
    int driverVersion = 0;

    cr = cudaDriverGetVersion(&driverVersion);
    CHECK_CUDA_ERROR

    printf("CUDA Driver Version: %d\n", driverVersion);
    if (driverVersion < 3010) {
        printf("Waiving: CUDA driver too old to support VDPAU interop.\n");
        common_exit_ok();
        return 0;
    }

    int ret;

    app_name = "vdpau_traceplay_cuda_toolkit";
    support_opt_gold_surface_type_mask =
        (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER) |
        (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER) |
        (1 << SURFACE_TYPE_VDP_OUT) |
        (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID) |
        (1 << SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_OUT);

    common_init();

    argc--;
    argv++;
    while (argc) {
        if ((ret = common_parse_arg(&argc, &argv)) != COMMON_PARSE_ARG_NOT_PARSED) {
            if (ret != COMMON_PARSE_ARG_OK) {
                help();
                common_exit_error();
            }
        }
        else {
            printf("ERROR: Unknown option '%s'\n", argv[0]);
            help();
            common_exit_error();
        }
        argc--;
        argv++;
    }

    if (common_check_args() != COMMON_CHECK_ARGS_OK) {
        help();
        common_exit_error();
    }

    common_load_stream();

    initX11();
    initVDPAU();
    common_playback_init(
        vdp_device,
        vdp_get_proc_address,
        (opt_gold_surface_type == SURFACE_TYPE_CUDA_TOOLKIT_INTEROP_VID)
    );
    struct Traceplay_Ctx *traceplay_ctx = traceplay_init(
        vdp_device,
        vdp_get_proc_address,
        TRACEPLAY_PLAY_TYPE_CUSTOM_MIXER,
        &tpcb,
        trace_stream_conf_ptr
    );
    CHECK_PTR(traceplay_ctx)
    traceplay_play(
        traceplay_ctx,
        trace_cmds_ptr,
        opt_repeat_count
    );
    traceplay_fini(
        traceplay_ctx
    );
    common_playback_fini();
    finiCUDA();
    finiVDPAU();
    finiX11();

    common_exit_ok();
    return 0;
}

